/* Audio Library for Teensy 3.X
 * Copyright (c) 2014, Paul Stoffregen, paul@pjrc.com
 *
 * Development of this audio library was funded by PJRC.COM, LLC by sales of
 * Teensy and Audio Adaptor boards.  Please support PJRC's efforts to develop
 * open source software by purchasing Teensy or other PJRC products.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice, development funding notice, and this permission
 * notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
 
 /* 
 *  Extended by Chip Audette, OpenAudio, May 2019
 *  Converted to F32 and to variable audio block length
 *	The F32 conversion is under the MIT License.  Use at your own risk.
 */
 
#ifndef output_i2s_quad_f32_h_
#define output_i2s_quad_f32_h_

#include <Arduino.h>
#include <arm_math.h>  //to define float32_t
#include "AudioStream_F32.h"
//include "AudioStream.h"
#include "DMAChannel.h"
#include <arm_math.h>
#include "output_i2s_quad_F32.h" //for AudioInputI2SQuad_F32()
#include "output_i2s_F32.h"  //for scale_f32_to_i16()


class AudioOutputI2SQuad_F32 : public AudioStream_F32
{
public:
	AudioOutputI2SQuad_F32(void) : AudioStream_F32(4, inputQueueArray) { 
		audio_block_samples = MAX_AUDIO_BLOCK_SAMPLES_F32; //use default
		begin();
	}
	AudioOutputI2SQuad_F32(const AudioSettings_F32 &settings) : AudioStream_F32(4, inputQueueArray)
	{ 
		sample_rate_Hz = settings.sample_rate_Hz;
		audio_block_samples = settings.audio_block_samples;
		begin(); 	
	}
	AudioOutputI2SQuad_F32(const AudioSettings_F32 &settings, uint32_t *tx_buff) : AudioStream_F32(4, inputQueueArray) { 
		sample_rate_Hz = settings.sample_rate_Hz;
		audio_block_samples = settings.audio_block_samples;
		i2s_tx_buffer = tx_buff;
		zerodata = new float32_t[settings.audio_block_samples/2]{0}; //Need zeros for half an audio block, just 1 channel, init to zero
		begin(); 
	}
	virtual void update(void);
	void begin(void);
	friend class AudioInputI2SQuad_F32;
	//static void scale_f32_to_i16( float32_t *p_f32, float32_t *p_i16, int len) ;
	//static void scale_f32_to_i24( float32_t *p_f32, float32_t *p_i16, int len) ;
	//static void scale_f32_to_i32( float32_t *p_f32, float32_t *p_i32, int len) ;
	static uint32_t *i2s_tx_buffer;
protected: 
	static void config_i2s(void);
	static audio_block_f32_t *block_ch1_1st;
	static audio_block_f32_t *block_ch2_1st;
	static audio_block_f32_t *block_ch3_1st;
	static audio_block_f32_t *block_ch4_1st;
	static bool update_responsibility;
	static DMAChannel dma;
	static void isr(void);
	static void isr_shuffleDataBlocks(audio_block_f32_t *&, audio_block_f32_t *&, uint32_t &);
	void update_1chan(const int, audio_block_f32_t *&, audio_block_f32_t *&, uint32_t &);
private:
	static audio_block_f32_t *block_ch1_2nd;
	static audio_block_f32_t *block_ch2_2nd;
	static audio_block_f32_t *block_ch3_2nd;
	static audio_block_f32_t *block_ch4_2nd;
	static uint32_t ch1_offset;
	static uint32_t ch2_offset;
	static uint32_t ch3_offset;
	static uint32_t ch4_offset;
	audio_block_f32_t *inputQueueArray[4];
	static float sample_rate_Hz;
	static int audio_block_samples;
	volatile uint8_t enabled = 1;
	static float32_t *zerodata;

};

#endif
